------------------------------------------------------------------------
-- General lazy vectors
------------------------------------------------------------------------

module LazyVec.General where

open import Relation.Binary.PropositionalEquality
open import Relation.Binary.PropositionalEquality1
open import Relation.Binary.HeterogeneousEquality
open import Data.Function
open import Data.Product
open import Data.Unit
open import Data.Empty
open import Data.Nat
open import Data.Vec hiding (_>>=_)

open import Thunk
open import LazyVec.VecSubtract
open import LazyVec.General.Lemmas

------------------------------------------------------------------------
-- Lazy vectors

-- Cost vectors.

CostVec : ℕ -> Set
CostVec = Vec ℕ

-- Lazy vectors defined using recursion.

VecL : (a : Set) (n : ℕ) -> CostVec n -> Set
VecL a ._ []       = ⊤
VecL a ._ (c ∷ cs) = a × Thunk c (VecL a _ cs)

------------------------------------------------------------------------
-- Paying off debits inside a lazy vector

-- Note that all the pay functions below are obviously correct, since
-- they return the input vector unchanged.

abstract

  -- Paying off at a certian location.

  payAt :  forall {a n₁ n₂} {cs₁ : CostVec n₁} {cs₂ : CostVec n₂}
                  {c} c'
        -> VecL a (n₁ + (1 + n₂)) (cs₁ ++ c ∷ cs₂)
        -> Thunk (1 + c')
                 (VecL a (n₁ + (1 + n₂)) (cs₁ ++ (c ∸ c') ∷ cs₂))
  payAt {a} {n₁} {n₂} {cs₁} {cs₂} {c} c' xs = √
    ≡₁-subst (≡₁-cong (Thunk c') (prop₂ cs₁))
             (pay (C cs₁) c' (≡₁-subst (prop₁ cs₁) xs))
    where
    C : forall {n} -> CostVec n -> Ctxt
    C []       = a ×• •
    C (c ∷ cs) = a ×• Thunk• c (C cs)

    prop₁ :  forall {n₁} (cs₁ : CostVec n₁)
          -> VecL a (n₁ + (1 + n₂)) (cs₁ ++ c ∷ cs₂) ≡₁
             C cs₁ [ Thunk c (VecL a n₂ cs₂)  ]
    prop₁ []       = ≡₁-refl
    prop₁ (c ∷ cs) = ≡₁-cong (\t -> a × Thunk c t) (prop₁ cs)

    prop₂ :  forall {n₁} (cs₁ : CostVec n₁)
          -> C cs₁ [ Thunk (c ∸ c') (VecL a n₂ cs₂) ] ≡₁
             VecL a (n₁ + (1 + n₂)) (cs₁ ++ (c ∸ c') ∷ cs₂)
    prop₂ []       = ≡₁-refl
    prop₂ (c ∷ cs) = ≡₁-cong (\t -> a × Thunk c t) (prop₂ cs)

abstract

  -- Paying off the first tick found.

  payL₁ :  forall {a n} cs
        -> VecL a n cs -> Thunk 2 (VecL a n (cs ⊖ 1))
  payL₁ {a = a} cs xs = √
    ≡₁-subst (≡₁-cong (Thunk 1) (prop₂ cs))
             (pay (C a cs) 1 (≡₁-subst (prop₁ cs) xs))
    where
    C : Set -> forall {n} -> CostVec n -> Ctxt
    C a []           = const• ⊤
    C a (zero  ∷ cs) = a ×• Thunk• zero (C a cs)
    C a (suc c ∷ cs) = a ×• •

    D : Set -> forall {n} -> CostVec n -> Set
    D a []           = ⊥
    D a (zero  ∷ cs) = D a cs
    D a (suc c ∷ cs) = VecL a _ cs

    N : forall {n} -> CostVec n -> ℕ
    N []           = zero
    N (zero  ∷ cs) = N cs
    N (suc c ∷ cs) = suc c

    prop₁ :  forall {a n} cs
          -> VecL a n cs ≡₁ C a cs [ Thunk (N cs) (D a cs) ]
    prop₁     []           = ≡₁-refl
    prop₁ {a} (zero  ∷ cs) = ≡₁-cong (\t -> a × Thunk 0 t) (prop₁ cs)
    prop₁     (suc c ∷ cs) = ≡₁-refl

    prop₂ :  forall {a n} cs
          -> C a cs [ Thunk (N cs ∸ 1) (D a cs) ] ≡₁ VecL a n (cs ⊖ 1)
    prop₂     []           = ≡₁-refl
    prop₂ {a} (zero  ∷ cs) = ≡₁-cong (\t -> a × Thunk 0 t) (prop₂ cs)
    prop₂ {a} (suc c ∷ cs) = ≡₀₁-cong (\cs -> a × Thunk c (VecL a _ cs))
                                      (≡-sym $ xs⊖0∸x≡xs c cs)

abstract

  -- Paying off the first m ticks found.

  payL :  forall {a n} cs
       -> (m : ℕ) -> VecL a n cs
       -> Thunk (1 + 3 * m) (VecL a n (cs ⊖ m))
  payL {a = a} cs zero    xs = √
    return (≡-subst (VecL a _) (≡-sym $ xs⊖0≡xs cs) xs)
  payL {a = a} cs (suc m) xs = √
    cast (lem₁ m)
      (≡-subst (\cs -> Thunk (3 + 3 * m) (VecL a _ cs))
               (xs⊖1⊖x≡xs⊖1+x m cs) $
         payL (cs ⊖ 1) m =<< payL₁ cs xs)


------------------------------------------------------------------------
-- Casts

castVecL
  :  forall {a m₁ m₂ n₁ n₂ cs₁ cs₂}
  -> m₁ ≡ m₂ -> n₁ ≡ n₂ -> cs₁ ≅ cs₂
  -> Thunk m₁ (VecL a n₁ cs₁) -> Thunk m₂ (VecL a n₂ cs₂)
castVecL ≡-refl ≡-refl ≅-refl xs = xs

------------------------------------------------------------------------
-- Rotation

-- rotate xs ys zs is basically xs ++ reverse ys ++ zs.

rotate
  :  forall {a n₁ n₂} cs₁ cs₂
  -> VecL a n₁ cs₁ -> Vec a (1 + n₁) -> VecL a n₂ cs₂
  -> Thunk 1 (VecL a (n₁ + (1 + n₁ + n₂))
                     (map (\c -> c + 2) cs₁
                      ++ replicate {n = 1 + n₁} 0
                      ++ cs₂))
rotate {n₁ = zero} [] _ tt (y ∷ []) zs = √ return (y , return zs)
rotate {n₁ = suc n₁} {n₂ = n₂} (c ∷ cs₁) cs₂
       (x , xs) (y ∷ ys) zs = √ return (x ,
  castVecL ≡-refl (lem₂ n₁ n₂) (lem₃ (map (\c -> c + 2) cs₁) n₁ 0 cs₂)
  (xs >>= \xs' -> √
   rotate cs₁ (0 ∷ cs₂) xs' ys (y , return zs)))

-- rotate₀ xs ys is basically xs ++ reverse ys.

rotate₀
  :  forall {a n₁} cs₁
  -> VecL a n₁ cs₁ -> Vec a (1 + n₁)
  -> Thunk 2 (VecL a (n₁ + (1 + n₁))
                     (map (\c -> c + 2) cs₁ ++ replicate 0))
rotate₀ {n₁ = n₁} cs₁ xs ys = √
  castVecL ≡-refl (lem₄ n₁)
                  (lem₅ (map (\c -> c + 2) cs₁)
                        (replicate {n = 1 + n₁} 0))
    (rotate cs₁ [] xs ys tt)
