------------------------------------------------------------------------
-- Simple lazy vectors
------------------------------------------------------------------------

-- Lazy vectors with a fixed cost per tail.

{-# OPTIONS --no-positivity-check #-}

-- Thunk is abstract, so the positivity checker cannot "see" that it
-- is just the identity. If Thunk were inlined the VecL type would be
-- strictly positive.

module LazyVec.Simple where

open import Data.Nat
open import Data.Nat.Properties
open ℕ-semiringSolver
open import Data.Fin
open import Data.Vec using ([]; _∷_)
open import Relation.Binary.PropositionalEquality

open import Thunk

------------------------------------------------------------------------
-- Lazy vectors

infixr 5 _∷L_

data VecL (a : Set) (c : ℕ) : ℕ -> Set where
  []L  :  VecL a c 0
  _∷L_ :  forall {n}
       -> a -> Thunk c (VecL a c n) -> VecL a c (1 + n)

------------------------------------------------------------------------
-- Boring lemmas

private
 abstract
   lem₁ : forall c n -> c + (1 + n * (1 + c)) ≡ (1 + n) * (1 + c)
   lem₁ c n = prove (n ∷ c ∷ [])
                   (C :+ (con 1 :+ N :* (con 1 :+ C)))
                   ((con 1 :+ N) :* (con 1 :+ C))
                   ≡-refl
     where
     N = var zero
     C = var (suc zero)

   lem₂ : forall k c -> c + k ≡ k + c
   lem₂ k c = prove (k ∷ c ∷ []) (C :+ K) (K :+ C) ≡-refl
     where K = var zero; C = var (suc zero)

   lem₃ : forall c c' -> 1 + c + (c' + 1) ≡ 2 + c + c'
   lem₃ c c' = prove (c ∷ c' ∷ [])
               (con 1 :+ C :+ (C' :+ con 1))
               (con 2 :+ C :+ C')
               ≡-refl
     where
     C  = var zero
     C' = var (suc zero)

------------------------------------------------------------------------
-- Extraction functions

head : forall {a c n} -> VecL a c (1 + n) -> Thunk 1 a
head (x ∷L xs) = √ return x

tail : forall {a c n} -> VecL a c (1 + n) -> Thunk (1 + c) (VecL a c n)
tail (x ∷L xs) = √ xs

last :  forall {a c n}
     -> VecL a c (1 + n) -> Thunk (1 + (1 + n) * (1 + c)) a
last {a} {c} {n} (x ∷L xs) = √ cast (lem₁ c n) (last' x =<< xs)
  where
  last' : forall {n} -> a -> VecL a c n -> Thunk (1 + n * (1 + c)) a
  last'         x []L       = √ return x
  last' {suc n} x (y ∷L ys) = √ cast (lem₁ c n) (last' y =<< ys)

------------------------------------------------------------------------
-- Functions for increasing the cost index

waitL₁ :  forall {c n a} -> VecL a c n
      -> Thunk 1 (VecL a (1 + c) n)
waitL₁             []L        = √ return []L
waitL₁ {c} {suc _} (x ∷L xs)  = √
  return (x ∷L cast (lem₂ 1 c) (waitL₁ =<< xs))

waitL :  forall {c' n a} c -> VecL a c' n
      -> Thunk 1 (VecL a (2 + c + c') n)
waitL              c []L        = √ return []L
waitL {c'} {suc _} c (x ∷L xs)  = √
  return (x ∷L cast (lem₃ c c') (wait c (waitL c =<< xs)))

------------------------------------------------------------------------
-- General-purpose functions

infixr 5 _++L_

_++L_ : forall {a c m n} ->
        VecL a c m -> VecL a c n -> Thunk 2 (VecL a (3 + c) (m + n))
_++L_         []L       ys = √ waitL 1 ys
_++L_ {c = c} (x ∷L xs) ys = √ returnW _
  (x ∷L cast (lem₂ 3 c) (xs >>= \xs' -> √ xs' ++L ys))
