------------------------------------------------------------------------
-- Statements of the main results
------------------------------------------------------------------------

-- Some of these results are proved elsewhere (see the module Proofs).

module MainResults where

------------------------------------------------------------------------
-- The type system is well-defined, as are the simple and thunked
-- semantics

-- The following modules define the type system and the semantics,
-- which by construction are well-defined (well-typed etc.).

open import Kind
open import TypeSystem
open import Erasure
open import Thunked
open import Substitution
open import Value
import Heap
import Heap.Erasure
open import OperationalSemantics.Binding
open import OperationalSemantics.Simple
open import PayCtxt
open import TypeTriple
open import OperationalSemantics.GPay
open import OperationalSemantics.Thunked

------------------------------------------------------------------------
-- The simple semantics is functional

-- To state functionality we first need to define equality between the
-- results of derivations.

open VarSubst
open Heap↦
import Data.Nat
open Data.Nat using (ℕ)
import Relation.Binary.PropositionalEquality
open Relation.Binary.PropositionalEquality using (_≡_)
open import Relation.Binary.HeterogeneousEquality using (_≅_)

data _=⇓=_
  {Γ : _} {τ : _} {Σ₁ : Heap Γ} {t : Γ ⊢ τ}
  {Δ¹ : _} {ρ¹ : Γ ⇒ Δ¹} {Σ₂¹ : Heap Δ¹} {v¹ : Δ¹ ⊢⇓ τ} {n¹ : ℕ}
  {Δ² : _} {ρ² : Γ ⇒ Δ²} {Σ₂² : Heap Δ²} {v² : Δ² ⊢⇓ τ} {n² : ℕ}
  (d¹ : Σ₁ ∣ t ⇓ ρ¹ ⋉ Σ₂¹ ∣ v¹ √ n¹)
  (d² : Σ₁ ∣ t ⇓ ρ² ⋉ Σ₂² ∣ v² √ n²)
  : Set where
  results-= : Δ¹ ≅ Δ² -> ρ¹ ≅ ρ² -> Σ₂¹ ≅ Σ₂² -> v¹ ≅ v² -> d¹ =⇓= d²

-- Now functionality can be stated.

Functionality : Set
Functionality =  forall {Γ : Ctxt simple} {τ Σ₁} {t : Γ ⊢ τ}
                        {Δ¹} {ρ¹ : Γ ⇒ Δ¹} {Σ₂¹ v¹ n¹}
                        {Δ²} {ρ² : Γ ⇒ Δ²} {Σ₂² v² n²}
              -> (d¹ : Σ₁ ∣ t ⇓ ρ¹ ⋉ Σ₂¹ ∣ v¹ √ n¹)
              -> (d² : Σ₁ ∣ t ⇓ ρ² ⋉ Σ₂² ∣ v² √ n²)
              -> d¹ =⇓= d²

------------------------------------------------------------------------
-- The thunked semantics is sound and complete with respect to the
-- simple one

open Heap↦Erasure
open Relation.Binary.PropositionalEquality using (≡-subst)

Soundness : Set
Soundness =  forall {Γ Δ τ Σ₁ ρ Σ₂ n} {e : Γ ⊢ τ} {v : Δ ⊢⇓ τ}
          -> Σ₁ ∣ e ⇊ ρ ⋉ Σ₂ ∣ v √ n
          -> ⌜ Σ₁ ⌝ʰ ∣ ⌜ e ⌝ ⇓ ⌜ ρ ⌝⇒ ⋉ ⌜ Σ₂ ⌝ʰ ∣ ⌜ v ⌝⇓ √ n

-- The type ∃Thunked Σ₁ e ρ Σ₂ v n, if inhabited, says that Σ₁ | e
-- evaluates in n steps (in the thunked semantics) to something which
-- erases to Σ₂ | v. (The statement is actually a bit stronger, since
-- the resulting substitution also has to be related to ρ.)

data ∃Thunked {Γ : Ctxt thunked} {Δ : Ctxt simple} {τ : Ty thunked}
              (Σ₁ : Heap Γ) (e : Γ ⊢ τ)
              (ρ : ⌜ Γ ⌝▻ ⇒ Δ) (Σ₂ : Heap Δ) (v : Δ ⊢⇓ ⌜ τ ⌝⋆) (n : ℕ)
              : Set where
  ∃thunked :  (Δ' : Ctxt thunked)
           -> (Δ-eq : Δ ≡ ⌜ Δ' ⌝▻)
           -> (ρ' : Γ ⇒ Δ')
           -> ≡-subst (_⇒_ _) Δ-eq ρ ≡ ⌜ ρ' ⌝⇒
           -> (Σ₂' : Heap Δ')
           -> ≡-subst Heap Δ-eq Σ₂ ≡ ⌜ Σ₂' ⌝ʰ
           -> (v' : Δ' ⊢⇓ τ)
           -> ≡-subst (\Δ -> Δ ⊢⇓ ⌜ τ ⌝⋆) Δ-eq v ≡ ⌜ v' ⌝⇓
           -> Σ₁ ∣ e ⇊ ρ' ⋉ Σ₂' ∣ v' √ n
           -> ∃Thunked Σ₁ e ρ Σ₂ v n

Completeness : Set
Completeness =  forall Γ {Δ τ} Σ₁ {ρ Σ₂ n} (e : Γ ⊢ τ) {v : Δ ⊢⇓ ⌜ τ ⌝⋆}
             -> ⌜ Σ₁ ⌝ʰ ∣ ⌜ e ⌝ ⇓ ρ ⋉ Σ₂ ∣ v √ n
             -> ∃Thunked Σ₁ e ρ Σ₂ v n

------------------------------------------------------------------------
-- The time bounds asserted by the type system are correct

open Data.Nat using (zero; _+_; _≤_)
open import Data.List using (sum)

-- A term of type Thunk n₁ (Thunk n₂ (Thunk n₃ Nat)) takes at most
-- n₁ + n₂ + n₃ steps amortised time to evaluate to WHNF. The time
-- function calculates this number, given a type.

time : Ty thunked -> ℕ
time (Thunk n τ) = n + time τ
time _           = zero

-- Heap bindings carry credit, corresponding to evaluation steps which
-- have already been paid off, but not yet performed.

credit↦ : forall {Γ : Ctxt thunked} {τ} -> Γ ↦ τ -> ℕ
credit↦ (↦ⁿˢ ns e) = sum ns

-- A heap inherits the credit of its bindings.

credit : forall {Γ : Ctxt thunked} -> Heap Γ -> ℕ
credit ∅ʰ       = zero
credit (Σ ▷ʰ b) = credit↦ b + credit Σ

-- Correctness statement.

Correctness : Set
Correctness =  forall {Γ Δ τ ρ Σ₁ Σ₂ n} {e : Γ ⊢ τ} {v : Δ ⊢⇓ τ}
            -> Σ₁ ∣ e ⇊ ρ ⋉ Σ₂ ∣ v √ n
            -> n + credit Σ₂ ≤ time τ + credit Σ₁

-- The statement above can be simplified. The simplified variant does
-- not make use of the thunked semantics, so it is easier to
-- understand. However, it is also less modular: no credit is attached
-- to Σ₂, so instances of this result cannot (easily) be chained
-- together.

Correctness' : Set
Correctness' =  forall {Γ Δ τ ρ} (Σ₁ : Heap Γ) {Σ₂ n}
                       (e : Γ ⊢ τ) {v : Δ ⊢⇓ ⌜ τ ⌝⋆}
             -> ⌜ Σ₁ ⌝ʰ ∣ ⌜ e ⌝ ⇓ ρ ⋉ Σ₂ ∣ v √ n
             -> n ≤ time τ + credit Σ₁
