------------------------------------------------------------------------
-- Heap bindings
------------------------------------------------------------------------

module OperationalSemantics.Binding where

open import Kind
open import TypeSystem
open import Erasure
open import Substitution
open VarSubst
open import Value
import Heap
import Heap.Erasure

open import Data.Nat using (ℕ; _∸_)
open import Data.List using ([]; _∷_; foldr) renaming ([_] to List)
open import Relation.Binary.PropositionalEquality
  using (_≡_; ≡-refl; ≡-subst; ≡-cong; ≡-sym)

infixl 48 _/↦_
infix  30 ↦_ ↦ⁿ_
infix  10 _↦_

------------------------------------------------------------------------
-- Multiple applications of Thunk

-- Thunks (n₁ ∷ n₂ ∷ []) τ ≡ Thunk n₁ (Thunk n₂ τ).

Thunks : List ℕ -> Ty thunked -> Ty thunked
Thunks ns τ = foldr Thunk τ ns

Thunks-lemma : forall ns {τ} -> ⌜ Thunks ns τ ⌝⋆ ≡ ⌜ τ ⌝⋆
Thunks-lemma []       = ≡-refl
Thunks-lemma (n ∷ ns) = Thunks-lemma ns

-- Generalised variant which vanishes in the simple case.

Thunks? : forall {k} -> Maybe (List ℕ) k -> Ty k -> Ty k
Thunks? {k = simple}  ns τ = τ
Thunks? {k = thunked} ns τ = Thunks ns τ

-- Thunks● (n₁ ∷ n₂ ∷ []) C ≡ Thunk● n₁ (Thunk● n₂ C).

Thunks● : List ℕ -> PayCtxt -> PayCtxt
Thunks● ns C = foldr Thunk● C ns

Thunks●-lemma :  forall ns C {τ}
              -> Thunks ns (C [ τ ]) ≡ Thunks● ns C [ τ ]
Thunks●-lemma []       C = ≡-refl
Thunks●-lemma (n ∷ ns) C = ≡-cong (Thunk n) (Thunks●-lemma ns C)

-- Removes one application of return⋆⇓ for every element in ns.

drop-return⋆⇓ :  forall {Γ τ} ns
              -> Γ ⊢⇓ Thunks ns τ -> Γ ⊢⇓ τ
drop-return⋆⇓ []       v            = v
drop-return⋆⇓ (n ∷ ns) (return⋆⇓ v) = drop-return⋆⇓ ns v

------------------------------------------------------------------------
-- Heap bindings

-- The heap bindings carry credit (in "ns"; see MainResult),
-- corresponding to terms which have been paid off but not yet
-- evaluated. The type of the binding reflects this credit.

data _↦_ : TmLike where
  ↦ⁿˢ :  forall {k} {Γ : Ctxt k} {τ} ns
      -> Γ ⊢ Thunks? ns τ -> Γ ↦ τ

-- Some special cases (these two cases could replace ↦ⁿˢ if there was
-- no generalised pay):

-- Ordinary binding.

↦_ : forall {k} {Γ : Ctxt k} {τ} -> Γ ⊢ τ -> Γ ↦ τ
↦_ {k = simple}  e = ↦ⁿˢ _  e
↦_ {k = thunked} e = ↦ⁿˢ [] e

-- A binding which has been (partially) paid off.

↦ⁿ_ : forall {Γ τ n} -> Γ ⊢ Thunk n τ -> Γ ↦ τ
↦ⁿ_ {n = n} e = ↦ⁿˢ (n ∷ []) e

-- Erases the heap bindings.

⌜_⌝↦ : TmEraser _↦_
⌜ ↦ⁿˢ ns e ⌝↦ = ↦ ≡-subst (_⊢_ _) (Thunks-lemma ns) ⌜ e ⌝

open module Heap↦ = Heap _↦_
module Heap↦Erasure = Heap.Erasure _↦_ ⌜_⌝↦

-- Variable substitutions can be applied to heap bindings.

_/↦_ : forall {k} {Γ Δ : Ctxt k} {τ} -> Γ ↦ τ -> Γ ⇒ Δ -> Δ ↦ τ
↦ⁿˢ ns e /↦ ρ = ↦ⁿˢ ns (e /⊢ ρ)

-- Paying off bindings.

gpay↦ :  forall {Γ σ n} C m
      -> Γ ↦ C [ Thunk n σ ] -> Γ ↦ C [ Thunk (n ∸ m) σ ]
gpay↦ C m (↦ⁿˢ ns e) =
  ↦ⁿˢ (m ∷ ns) (cast₁ (gpay (Thunks● ns C) m (cast₂ e)))
  where
  cast₁ = ≡-subst (_⊢_ _) (≡-sym (Thunks●-lemma (m ∷ ns) C))
  cast₂ = ≡-subst (_⊢_ _) (Thunks●-lemma ns C)
