------------------------------------------------------------------------
-- Functions for calculating the depth of a derivation tree
------------------------------------------------------------------------

-- The completeness proof in OperationalSemantics.Complete is by
-- lexicographic induction over this depth and another measure.

module OperationalSemantics.Depth where

open import TypeSystem
open import Substitution
open VarSubst
open import OperationalSemantics.Simple
open import Value
open import Relation.Binary.PropositionalEquality

-- The depth is not a natural number. The following type is more
-- convenient if we want to perform induction over the depth of a
-- derivation.

data Depth : Set where
  [0]     : Depth
  [suc]   : Depth -> Depth
  -- The following constructor really stands for "1 + the maximum of
  -- the two children".
  _[max]_ : Depth -> Depth -> Depth

mutual

  -- Calculates the depth.

  depth :  forall {Γ Δ τ Σ₁} {t : Γ ⊢ τ}
                  {ρ : Γ ⇒ Δ} {Σ₂ v n}
        -> Σ₁ ∣ t ⇓ ρ ⋉ Σ₂ ∣ v √ n -> Depth
  depth eval-λ           = [0]
  depth eval-λ²          = [0]
  depth eval-z           = [0]
  depth eval-s           = [0]
  depth eval-natrec      = [0]
  depth eval-,           = [0]
  depth eval-uncurry     = [0]
  depth (eval-vz d)      = [suc] (depth d)
  depth (eval-vs d)      = [suc] (depth d)
  depth (eval-fix d)     = [suc] (depth d)
  depth (eval-· d₁ d₂)   = depth d₁ [max] depth• d₂

  depth• :  forall {Γ Δ σ τ Σ₁} {v' : Γ ⊢⇓ σ → τ}
                   {ρ : Γ ▻ σ ⇒ Δ} {Σ₂ v n}
         -> Σ₁ ∣ v' •⇓ ρ ⋉ Σ₂ ∣ v √ n -> Depth
  depth• eval-·-s                   = [0]
  depth• (eval-·-λ d)               = [suc] (depth d)
  depth• eval-·-λ²                  = [0]
  depth• (eval-·-natrec-zero d₁ d₂) = depth d₁ [max] depth d₂
  depth• (eval-·-natrec-suc d₁ d₂)  = depth d₁ [max] depth d₂
  depth• (eval-·-uncurry d₁ d₂)     = depth d₁ [max] depth d₂

-- Functions for simplifying equalities involving depths.

drop-[suc] : forall {d₁ d₂} -> [suc] d₁ ≡ [suc] d₂ -> d₁ ≡ d₂
drop-[suc] ≡-refl = ≡-refl

drop-[max]d :  forall {d₁¹ d₂¹ d₁² d₂²}
          -> d₁¹ [max] d₂¹ ≡ d₁² [max] d₂² -> d₁¹ ≡ d₁²
drop-[max]d ≡-refl = ≡-refl

drop-d[max] :  forall {d₁¹ d₂¹ d₁² d₂²}
          -> d₁¹ [max] d₂¹ ≡ d₁² [max] d₂² -> d₂¹ ≡ d₂²
drop-d[max] ≡-refl = ≡-refl
