------------------------------------------------------------------------
-- Lazy operational semantics
------------------------------------------------------------------------

-- The operational semantics is inspired by Lanchbury's natural
-- semantics for lazy evaluation.

-- The evaluation relation is annotated by the number of steps
-- necessary to evaluate a term to weak head normal form. Note that
-- reindexing of de Bruijn indices is not counted, since this is not
-- typically part of efficient implementations. Similarly,
-- substituting a variable for another, adding heap bindings, or
-- looking up variables are not counted.

-- Note that the type of the relation ensures that the semantics is
-- well-scoped and well-typed, and that the result is in weak head
-- normal form.

-- The semantics is indexed on a substitution (of type Γ ⇒ Δ). This is
-- a variable substitution, i.e. a renaming. The resulting heap may
-- have more bindings than the initial one, and this substitution
-- describes how one can "weaken" the initial context to arrive at the
-- final one.

-- The inference rules in comments below describe the rules using
-- ordinary variables. They assume that all variables are suitably
-- fresh.

module OperationalSemantics.Simple where

open import Kind
open import TypeSystem
open import Substitution
open VarSubst
open import Value
open import OperationalSemantics.Binding
open Heap↦

open import Data.Nat using (ℕ; zero; suc; _+_)

infix 10 _∣_⇓_⋉_∣_√_ _∣_•⇓_⋉_∣_√_

mutual

  data _∣_⇓_⋉_∣_√_
    :  forall {Γ Δ : Ctxt simple} {τ}
    -> Heap Γ -> Γ ⊢ τ
    -> Γ ⇒ Δ -> Heap Δ -> Δ ⊢⇓ τ
    -> ℕ
    -> Set where

    -- Values.

    -- —————————————————————————
    --  Σ | λx.t ⇓ Σ | λx.t √ 0

    eval-λ
      :  forall {Γ σ τ Σ} {e : Γ ▻ σ ⊢ τ}
      -> Σ ∣ λ e ⇓ idˢ ⋉ Σ ∣ λ⇓ e √ zero

    -- ———————————————————————————
    --  Σ | λxy.t ⇓ Σ | λxy.t √ 0

    eval-λ²
      :  forall {Γ σ₁ σ₂ τ Σ} {e : Γ ▻ σ₁ ▻ σ₂ ⊢ τ}
      -> Σ ∣ λ² e ⇓ idˢ ⋉ Σ ∣ λ²⇓ e √ zero

    -- ———————————————————
    --  Σ | z ⇓ Σ | z √ 0

    eval-z
      :  forall {Γ} {Σ : Heap Γ}
      -> Σ ∣ z ⇓ idˢ ⋉ Σ ∣ z⇓ √ zero

    -- ———————————————————
    --  Σ | s ⇓ Σ | s √ 0

    eval-s
      :  forall {Γ} {Σ : Heap Γ}
      -> Σ ∣ s ⇓ idˢ ⋉ Σ ∣ s⇓ √ zero

    -- ——————————————————————————————————————————————————————————————
    --  Σ | natrec t₁ (λxy.t₂) ⇓ Σ, x₁ ↦ t₁ | natrec x₁ (λxy.t₂) √ 0

    eval-natrec
      :  forall {Γ σ Σ} {e₁ : Γ ⊢ σ} {e₂}
      -> Σ ∣ natrec e₁ e₂ ⇓ wk ⋉ Σ ▷ʰ ↦ e₁
                          ∣ natrec⇓ vz (e₂ /⊢ wk ↑ ↑) √ zero

    -- —————————————————————————————————————————————————
    --  Σ | (t₁ , t₂) ⇓ Σ, x ↦ t₁, y ↦ t₂ | (x , y) √ 0

    eval-,
      :  forall {Γ σ₁ σ₂ Σ} {e₁ : Γ ⊢ σ₁} {e₂ : Γ ⊢ σ₂}
      -> Σ ∣ (e₁ , e₂) ⇓ wk ∘ˢ wk ⋉ Σ ▷ʰ ↦ e₁ ▷ʰ ↦ e₂ /⊢ wk
                       ∣ (vs vz ,⇓ vz) √ zero

    -- ———————————————————————————————————————————————
    --  Σ | uncurry (λxy.t) ⇓ Σ | uncurry (λxy.t) √ 0

    eval-uncurry
      :  forall {Γ σ₁ σ₂ τ Σ} {e : Γ ▻ σ₁ ▻ σ₂ ⊢ τ}
      -> Σ ∣ uncurry e ⇓ idˢ ⋉ Σ ∣ uncurry⇓ e √ zero

    -- Variables.

    --             Σ₁ | t ⇓ Σ₂            | v √ n
    -- ———————————————————————————————————————————
    --  Σ₁, x ↦ t, Σ' | x ⇓ Σ₂, x ↦ v, Σ' | v √ n

    eval-vz
      :  forall {Γ Δ τ n ρ Σ₁ Σ₂} {e₁ : Γ ⊢ τ} {v : Δ ⊢⇓ τ}
      -> Σ₁         ∣ e₁     ⇓ ρ   ⋉ Σ₂                ∣ v        √ n
      -> Σ₁ ▷ʰ ↦ e₁ ∣ var vz ⇓ ρ ↑ ⋉ Σ₂ ▷ʰ ↦ valToTm v ∣ v /⊢⇓ wk √ n

    eval-vs
      :  forall {Γ Δ σ τ n ρ Σ₁ Σ₂}
                {x : Γ ∋ τ} {b : Γ ↦ σ} {v : Δ ⊢⇓ _}
      -> Σ₁      ∣ var x      ⇓ ρ   ⋉ Σ₂           ∣ v        √ n
      -> Σ₁ ▷ʰ b ∣ var (vs x) ⇓ ρ ↑ ⋉ Σ₂ ▷ʰ b /↦ ρ ∣ v /⊢⇓ wk √ n

    -- Fixpoints.

    --  Σ₁, x ↦ fix (λx.t) | t          ⇓ Σ₂ | v √ n
    -- ——————————————————————————————————————————————————
    --  Σ₁                 | fix (λx.t) ⇓ Σ₂ | v √ 1 + n

    eval-fix
      :  forall {Γ Δ σ n Σ₁ Σ₂} {ρ : _ ⇒ Δ} {e : Γ ▻ σ ⊢ σ} {v}
      -> Σ₁ ▷ʰ ↦ fix e ∣ e     ⇓ ρ       ⋉ Σ₂ ∣ v √ n
      -> Σ₁            ∣ fix e ⇓ wk ∘ˢ ρ ⋉ Σ₂ ∣ v √ suc n

    -- Applications.

    --  Σ₁          | t₁      ⇓ Σ₂ | v₁ √ n₁
    --  Σ₂, x₂ ↦ t₂ | v₁ • x₂ ⇓ Σ₃ | v  √ n₂
    -- ———————————————————————————————————————————
    --  Σ₁          | t₁ · t₂ ⇓ Σ₂ | v  √ n₁ + n₂

    eval-·
      :  forall {Γ Δ X σ τ n₁ n₂ Σ₁ Σ₂ Σ₃}
                {ρ₁ : Γ ⇒ Δ} {ρ₂ : _ ⇒ X}
                {e₁ : Γ ⊢ σ → τ} {e₂ : Γ ⊢ σ} {v v₁}
      -> Σ₁               ∣ e₁      ⇓ ρ₁           ⋉ Σ₂ ∣ v₁ √ n₁
      -> Σ₂ ▷ʰ ↦ e₂ /⊢ ρ₁ ∣ v₁     •⇓ ρ₂           ⋉ Σ₃ ∣ v  √ n₂
      -> Σ₁               ∣ e₁ · e₂ ⇓ wk⇒ ρ₁ ∘ˢ ρ₂ ⋉ Σ₃ ∣ v  √ (n₁ + n₂)

  data _∣_•⇓_⋉_∣_√_
    :  forall {Γ Δ : Ctxt simple} {σ τ}
    -> Heap (Γ ▻ σ) -> Γ ⊢⇓ σ → τ
    -> Γ ▻ σ ⇒ Δ -> Heap Δ -> Δ ⊢⇓ τ
    -> ℕ
    -> Set where

    -- ———————————————————————————
    --  Σ | s • x₂ ⇓ Σ | s x₂ √ 0

    eval-·-s
      :  forall {Γ} {Σ : Heap (Γ ▻ Nat)}
      -> Σ ∣ s⇓ •⇓ idˢ ⋉ Σ ∣ s₁⇓ vz √ zero

    --  Σ₁ | t₁[x := x₂] ⇓ Σ₂ | v √ n
    -- ———————————————————————————————————
    --  Σ₁ | λx.t₁ • x₂  ⇓ Σ₂ | v √ 1 + n

    eval-·-λ
      :  forall {Γ Δ n σ τ Σ₁ Σ₂} {ρ : _ ⇒ Δ} {e₁ : Γ ▻ σ ⊢ τ} {v}
      -> Σ₁ ∣ e₁     ⇓ ρ ⋉ Σ₂ ∣ v √ n
      -> Σ₁ ∣ λ⇓ e₁ •⇓ ρ ⋉ Σ₂ ∣ v √ suc n

    -- Note that nothing is charged for the following application; only
    -- when all arguments have been supplied (and hence evaluation of
    -- the right hand side can commence) is something charged. If you do
    -- not agree with this cost model you should avoid partial
    -- applications.

    -- ————————————————————————————————————————————
    --  Σ₁ | λxy.t₁ • x₂ ⇓ Σ₁ | λy.t₁[x := x₂] √ 0

    eval-·-λ²
      :  forall {Γ σ₁ σ₂ τ Σ} {e₁ : Γ ▻ σ₁ ▻ σ₂ ⊢ τ}
      -> Σ ∣ λ²⇓ e₁ •⇓ idˢ ⋉ Σ ∣ λ⇓ e₁ √ zero

    --  Σ₁ | x₂                        ⇓ Σ₂ | z √ n₁
    --  Σ₂ | x₁¹                       ⇓ Σ₃ | v √ n₂
    -- ———————————————————————————————————————————————————————
    --  Σ₁ | natrec x₁¹ (λxy.t₁²) • x₂ ⇓ Σ₃ | v √ 1 + n₁ + n₂

    eval-·-natrec-zero
      :  forall {Γ₁ Γ₂ Γ₃ n₁ n₂ τ ρ₁ Σ₁ Σ₂ Σ₃}
                {ρ₂ : Γ₂ ⇒ Γ₃} {x₁¹ : Γ₁ ∋ τ} {e₁²} {v}
      -> let x₁¹' = var x₁¹ /⊢ wk ∘ˢ ρ₁
             ρ₃   = ρ₁ ∘ˢ ρ₂
         in
         Σ₁ ∣ var vz           ⇓ ρ₁ ⋉ Σ₂ ∣ z⇓ √ n₁
      -> Σ₂ ∣ x₁¹'             ⇓ ρ₂ ⋉ Σ₃ ∣ v  √ n₂
      -> Σ₁ ∣ natrec⇓ x₁¹ e₁² •⇓ ρ₃ ⋉ Σ₃ ∣ v  √ suc (n₁ + n₂)

    --  Σ₁  | x₂                        ⇓ Σ₂ | s x₃ √ n₁
    --  Σ₂, y ↦ natrec x₁¹ (λxy.t₁²) · x₃
    --      | t₁²[x := x₃]              ⇓ Σ₃ | v    √ n₂
    -- ———————————————————————————————————————————————————————————
    --  Σ₁  | natrec x₁¹ (λxy.t₁²) • x₂ ⇓ Σ₃ | v    √ 1 + n₁ + n₂

    eval-·-natrec-suc
      :  forall {Γ₁ Γ₂ Γ₃ n₁ n₂ τ ρ₁ ρ₂ Σ₁ Σ₂ Σ₃}
                {x₁¹ : Γ₁ ∋ τ} {e₁²}
                {x₂ : Γ₂ ∋ Nat} {v : Γ₃ ⊢⇓ τ}
      -> let Σ₂'  = Σ₂ ▷ʰ ↦ (natrec (var x₁¹) e₁² /⊢ (wk ∘ˢ ρ₁)) · var x₂
             e₁²' = e₁² /⊢ (wk ∘ˢ ρ₁) ↑ ↑ /⊢ sub x₂ ↑
             ρ₃   = wk⇒ ρ₁ ∘ˢ ρ₂
         in
         Σ₁  ∣ var vz           ⇓ ρ₁ ⋉ Σ₂ ∣ s₁⇓ x₂ √ n₁
      -> Σ₂' ∣ e₁²'             ⇓ ρ₂ ⋉ Σ₃ ∣ v      √ n₂
      -> Σ₁  ∣ natrec⇓ x₁¹ e₁² •⇓ ρ₃ ⋉ Σ₃ ∣ v      √ suc (n₁ + n₂)

    --  Σ₁ | x₂                    ⇓ Σ₂ | (x₃ , y₃) √ n₁
    --  Σ₂ | t₁[x := x₃, y := y₃]  ⇓ Σ₃ | v         √ n₂
    -- ——————————————————————————————————————————————————————————
    --  Σ₁ | uncurry (λxy.t₁) • x₂ ⇓ Σ₃ | v         √ 1 + n₁ + n₂

    eval-·-uncurry
      :  forall {Γ₁ Γ₂ Γ₃ n₁ n₂ σ₁ σ₂ τ ρ₁ ρ₂ Σ₁ Σ₂ Σ₃}
                {e₁ : Γ₁ ▻ σ₁ ▻ σ₂ ⊢ τ}
                {x₂ : Γ₂ ∋ σ₁} {y₂ : Γ₂ ∋ σ₂} {v : Γ₃ ⊢⇓ τ}
      -> let e₁' = e₁ /⊢ (wk ∘ˢ ρ₁) ↑ ↑ /⊢ sub x₂ ↑ /⊢ sub y₂ in
         Σ₁ ∣ var vz       ⇓ ρ₁       ⋉ Σ₂ ∣ (x₂ ,⇓ y₂) √ n₁
      -> Σ₂ ∣ e₁'          ⇓ ρ₂       ⋉ Σ₃ ∣ v          √ n₂
      -> Σ₁ ∣ uncurry⇓ e₁ •⇓ ρ₁ ∘ˢ ρ₂ ⋉ Σ₃ ∣ v          √ suc (n₁ + n₂)
