------------------------------------------------------------------------
-- Annotated lazy operational semantics
------------------------------------------------------------------------

-- The thunked semantics is very similar to the simple one. The main
-- difference is that it also handles the library primitives.

-- Note that in the thunked case the substitution which is an index to
-- the semantics can, in addition to renaming variables, also reduce
-- the tick counts of heap elements (due to the presence of gwait∋).
-- This is part of the scheme used to give a semantics for gpay, as
-- explained in OperationalSemantics.GPay.

-- Note also that it is possible to merge the simple and thunked
-- operational semantics, just as the simple terms and the thunked,
-- run-time terms were merged into a single definition. This is not
-- done here, for two reasons:
--
-- * I tried, and some proofs became more complicated.
--
-- * More importantly, now the simplified correctness result can be
--   stated without referring to the thunked semantics (see
--   MainResults).

module OperationalSemantics.Thunked where

open import Kind
open import TypeSystem
open import Substitution
open VarSubst
open import Value
open import OperationalSemantics.Binding
open Heap↦
open import OperationalSemantics.GPay

open import Data.Nat using (ℕ; zero; suc; _+_)

infix 10 _∣_⇊_⋉_∣_√_ _∣_•⇊_⋉_∣_√_

mutual

  data _∣_⇊_⋉_∣_√_
    :  forall {Γ Δ : Ctxt thunked} {τ}
    -> Heap Γ -> Γ ⊢ τ
    -> Γ ⇒ Δ -> Heap Δ -> Δ ⊢⇓ τ
    -> ℕ
    -> Set where

    -- Values.

    eval-λ
      :  forall {Γ σ n τ Σ} {e : Γ ▻ σ ⊢ Thunk n τ}
      -> Σ ∣ λ e ⇊ idˢ ⋉ Σ ∣ λ⇓ e √ zero

    eval-λ²
      :  forall {Γ σ₁ σ₂ n τ Σ} {e : Γ ▻ σ₁ ▻ σ₂ ⊢ Thunk n τ}
      -> Σ ∣ λ² e ⇊ idˢ ⋉ Σ ∣ λ²⇓ e √ zero

    eval-z
      :  forall {Γ} {Σ : Heap Γ}
      -> Σ ∣ z ⇊ idˢ ⋉ Σ ∣ z⇓ √ zero

    eval-s
      :  forall {Γ} {Σ : Heap Γ}
      -> Σ ∣ s ⇊ idˢ ⋉ Σ ∣ s⇓ √ zero

    eval-natrec
      :  forall {Γ σ n Σ} {e₁ : Γ ⊢ Thunk n σ} {e₂}
      -> Σ ∣ natrec e₁ e₂ ⇊ wk ⋉ Σ ▷ʰ ↦ e₁
                          ∣ natrec⇓ vz (e₂ /⊢ wk ↑ ↑) √ zero

    eval-,
      :  forall {Γ σ₁ σ₂ Σ} {e₁ : Γ ⊢ σ₁} {e₂ : Γ ⊢ σ₂}
      -> Σ ∣ (e₁ , e₂) ⇊ wk ∘ˢ wk ⋉ Σ ▷ʰ ↦ e₁ ▷ʰ ↦ e₂ /⊢ wk
                       ∣ (vs vz ,⇓ vz) √ zero

    eval-uncurry
      :  forall {Γ σ₁ σ₂ n τ Σ} {e : Γ ▻ σ₁ ▻ σ₂ ⊢ Thunk n τ}
      -> Σ ∣ uncurry e ⇊ idˢ ⋉ Σ ∣ uncurry⇓ e √ zero

    -- Variables.

    eval-vz
      :  forall {Γ Δ τ ms n ρ Σ₁ Σ₂}
                {e₁ : Γ ⊢ Thunks ms τ} {v : Δ ⊢⇓ Thunks ms τ}
      -> let v' = drop-return⋆⇓ ms v
             e₂ = valToTm v'
         in
         Σ₁              ∣ e₁     ⇊ ρ   ⋉ Σ₂         ∣ v         √ n
      -> Σ₁ ▷ʰ ↦ⁿˢ ms e₁ ∣ var vz ⇊ ρ ↑ ⋉ Σ₂ ▷ʰ ↦ e₂ ∣ v' /⊢⇓ wk √ n

    eval-vs
      :  forall {Γ Δ σ τ n ρ Σ₁ Σ₂}
                {x : Γ ∋ τ} {b : Γ ↦ σ} {v : Δ ⊢⇓ _}
      -> Σ₁      ∣ var x      ⇊ ρ   ⋉ Σ₂           ∣ v        √ n
      -> Σ₁ ▷ʰ b ∣ var (vs x) ⇊ ρ ↑ ⋉ Σ₂ ▷ʰ b /↦ ρ ∣ v /⊢⇓ wk √ n

    eval-gwait∋
      :  forall {Γ Δ τ m₁ m₂ n Σ₁ Σ₂ C} {ρ : Γ ⇒ Δ}
                {x : Γ ∋ C [ Thunk m₁ τ ]} {v}
      -> Σ₁ ∣ var x               ⇊ ρ ⋉ Σ₂ ∣ v             √ n
      -> Σ₁ ∣ var (gwait∋ C m₂ x) ⇊ ρ ⋉ Σ₂ ∣ gwait⇓ C m₂ v √ n

    -- Primitives.

    eval-return
      :  forall {Γ Δ τ n Σ₁ Σ₂} {ρ : Γ ⇒ Δ} {e : Γ ⊢ τ} {v}
      -> Σ₁ ∣ e        ⇊ ρ ⋉ Σ₂ ∣ v          √ n
      -> Σ₁ ∣ return e ⇊ ρ ⋉ Σ₂ ∣ return⋆⇓ v √ n

    eval-√
      :  forall {Γ Δ τ m n Σ₁ Σ₂} {ρ : Γ ⇒ Δ} {e : Γ ⊢ Thunk m τ} {v}
      -> Σ₁ ∣ e   ⇊ ρ ⋉ Σ₂ ∣ return⋆⇓ v √ n
      -> Σ₁ ∣ √ e ⇊ ρ ⋉ Σ₂ ∣ return⋆⇓ v √ n

    eval-pay
      :  forall {Γ Δ τ m₁ m₂ n Σ₁ Σ₂} {ρ : Γ ⇒ Δ}
                {e : Γ ⊢ Thunk m₁ τ} {v}
      -> Σ₁ ∣ e        ⇊ ρ ⋉ Σ₂ ∣ return⋆⇓ v            √ n
      -> Σ₁ ∣ pay m₂ e ⇊ ρ ⋉ Σ₂ ∣ return⋆⇓ (return⋆⇓ v) √ n

    eval-gpay
      :  forall {Γ Δ τ m₁ m₂ n Σ₁ Σ₂ C} {ρ₁ : Γ ⇒ Δ}
                {e : Γ ⊢ C [ Thunk m₁ τ ]} {v}
      -> let r  = gpayResult C m₂ Σ₂ v
             ρ₂ = ρ₁ ∘ˢ GPay.ρ r
         in
         Σ₁ ∣ e           ⇊ ρ₁ ⋉ Σ₂        ∣ v                    √ n
      -> Σ₁ ∣ gpay C m₂ e ⇊ ρ₂ ⋉ GPay.Σ₂ r ∣ return⋆⇓ (GPay.v₂ r) √ n

    eval->>=
      :  forall {Γ Δ X σ τ m₁ m₂ n₁ n₂ Σ₁ Σ₂ Σ₃} {ρ₁ : Γ ⇒ Δ} {ρ₂ : _ ⇒ X}
                {e₁ : Γ ⊢ Thunk m₁ σ} {e₂ : Γ ⊢ σ → Thunk m₂ τ} {v v₂}
      -> let ρ₃ = wk⇒ ρ₁ ∘ˢ ρ₂ in
         Σ₁                ∣ e₂        ⇊ ρ₁ ⋉ Σ₂ ∣ v₂         √ n₁
      -> Σ₂ ▷ʰ ↦ⁿ e₁ /⊢ ρ₁ ∣ v₂       •⇊ ρ₂ ⋉ Σ₃ ∣ return⋆⇓ v √ n₂
      -> Σ₁                ∣ e₁ >>= e₂ ⇊ ρ₃ ⋉ Σ₃ ∣ return⋆⇓ v √ (n₁ + n₂)

    eval-gwait
      :  forall {Γ Δ τ m₁ m₂ n Σ₁ Σ₂ C} {ρ : Γ ⇒ Δ}
                {e : Γ ⊢ C [ Thunk m₁ τ ]} {v}
      -> Σ₁ ∣ e            ⇊ ρ ⋉ Σ₂ ∣ v             √ n
      -> Σ₁ ∣ gwait C m₂ e ⇊ ρ ⋉ Σ₂ ∣ gwait⇓ C m₂ v √ n

    -- Fixpoints.

    eval-fix
      :  forall {Γ Δ σ m n Σ₁ Σ₂} {ρ : _ ⇒ Δ}
                {e : Γ ▻ Thunk (suc m) σ ⊢ Thunk m σ} {v}
      -> Σ₁ ▷ʰ ↦ fix e ∣ e     ⇊ ρ       ⋉ Σ₂ ∣ return⋆⇓ v √ n
      -> Σ₁            ∣ fix e ⇊ wk ∘ˢ ρ ⋉ Σ₂ ∣ return⋆⇓ v √ suc n

    -- Applications.

    eval-·
      :  forall {Γ Δ X σ τ n₁ n₂ Σ₁ Σ₂ Σ₃}
                {ρ₁ : Γ ⇒ Δ} {ρ₂ : _ ⇒ X}
                {e₁ : Γ ⊢ σ → τ} {e₂ : Γ ⊢ σ} {v v₁}
      -> Σ₁               ∣ e₁      ⇊ ρ₁           ⋉ Σ₂ ∣ v₁ √ n₁
      -> Σ₂ ▷ʰ ↦ e₂ /⊢ ρ₁ ∣ v₁     •⇊ ρ₂           ⋉ Σ₃ ∣ v  √ n₂
      -> Σ₁               ∣ e₁ · e₂ ⇊ wk⇒ ρ₁ ∘ˢ ρ₂ ⋉ Σ₃ ∣ v  √ (n₁ + n₂)

  data _∣_•⇊_⋉_∣_√_
    :  forall {Γ Δ : Ctxt thunked} {σ τ}
    -> Heap (Γ ▻ σ) -> Γ ⊢⇓ σ → τ
    -> Γ ▻ σ ⇒ Δ -> Heap Δ -> Δ ⊢⇓ τ
    -> ℕ
    -> Set where

    eval-·-s
      :  forall {Γ} {Σ : Heap (Γ ▻ Nat)}
      -> Σ ∣ s⇓ •⇊ idˢ ⋉ Σ ∣ s₁⇓ vz √ zero

    eval-·-λ
      :  forall {Γ Δ m n σ τ Σ₁ Σ₂} {ρ : _ ⇒ Δ}
                {e₁ : Γ ▻ σ ⊢ Thunk m τ} {v}
      -> Σ₁ ∣ e₁     ⇊ ρ ⋉ Σ₂ ∣ return⋆⇓ v √ n
      -> Σ₁ ∣ λ⇓ e₁ •⇊ ρ ⋉ Σ₂ ∣ return⋆⇓ v √ suc n

    eval-·-λ²
      :  forall {Γ m σ₁ σ₂ τ Σ}
                {e₁ : Γ ▻ σ₁ ▻ σ₂ ⊢ Thunk m τ}
      -> Σ ∣ λ²⇓ e₁ •⇊ idˢ ⋉ Σ ∣ λ⇓ e₁ √ zero

    eval-·-natrec-zero
      :  forall {Γ₁ Γ₂ Γ₃ m n₁ n₂ τ ρ₁ Σ₁ Σ₂ Σ₃}
                {ρ₂ : Γ₂ ⇒ Γ₃} {x₁¹ : Γ₁ ∋ Thunk m τ} {e₁²} {v}
      -> let x₁¹' = var x₁¹ /⊢ wk ∘ˢ ρ₁
             ρ₃   = ρ₁ ∘ˢ ρ₂
         in
         Σ₁ ∣ var vz           ⇊ ρ₁ ⋉ Σ₂ ∣ z⇓         √ n₁
      -> Σ₂ ∣ x₁¹'             ⇊ ρ₂ ⋉ Σ₃ ∣ return⋆⇓ v √ n₂
      -> Σ₁ ∣ natrec⇓ x₁¹ e₁² •⇊ ρ₃ ⋉ Σ₃ ∣ return⋆⇓ v √ suc (n₁ + n₂)

    eval-·-natrec-suc
      :  forall {Γ₁ Γ₂ Γ₃ m n₁ n₂ τ ρ₁ ρ₂ Σ₁ Σ₂ Σ₃}
                {x₁¹ : Γ₁ ∋ Thunk m τ} {e₁²}
                {x₂ : Γ₂ ∋ Nat} {v : Γ₃ ⊢⇓ τ}
      -> let Σ₂'  = Σ₂ ▷ʰ ↦ (natrec (var x₁¹) e₁² /⊢ (wk ∘ˢ ρ₁)) · var x₂
             e₁²' = e₁² /⊢ (wk ∘ˢ ρ₁) ↑ ↑ /⊢ sub x₂ ↑
             ρ₃   = wk⇒ ρ₁ ∘ˢ ρ₂
         in
         Σ₁  ∣ var vz           ⇊ ρ₁ ⋉ Σ₂ ∣ s₁⇓ x₂     √ n₁
      -> Σ₂' ∣ e₁²'             ⇊ ρ₂ ⋉ Σ₃ ∣ return⋆⇓ v √ n₂
      -> Σ₁  ∣ natrec⇓ x₁¹ e₁² •⇊ ρ₃ ⋉ Σ₃ ∣ return⋆⇓ v √ suc (n₁ + n₂)

    eval-·-uncurry
      :  forall {Γ₁ Γ₂ Γ₃ m n₁ n₂ σ₁ σ₂ τ ρ₁ ρ₂ Σ₁ Σ₂ Σ₃}
                {e₁ : Γ₁ ▻ σ₁ ▻ σ₂ ⊢ Thunk m τ}
                {x₂ : Γ₂ ∋ σ₁} {y₂ : Γ₂ ∋ σ₂} {v : Γ₃ ⊢⇓ τ}
      -> let e₁' = e₁ /⊢ (wk ∘ˢ ρ₁) ↑ ↑ /⊢ sub x₂ ↑ /⊢ sub y₂ in
         Σ₁ ∣ var vz       ⇊ ρ₁       ⋉ Σ₂ ∣ (x₂ ,⇓ y₂) √ n₁
      -> Σ₂ ∣ e₁'          ⇊ ρ₂       ⋉ Σ₃ ∣ return⋆⇓ v √ n₂
      -> Σ₁ ∣ uncurry⇓ e₁ •⇊ ρ₁ ∘ˢ ρ₂ ⋉ Σ₃ ∣ return⋆⇓ v √ suc (n₁ + n₂)
